/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007 Bull S.A.S.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:ClientClusterView.java 949 2007-06-02 17:24:33Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.carol.cmi.controller.client;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import net.jcip.annotations.ThreadSafe;

import org.ow2.carol.cmi.lb.data.LBPolicyData;
import org.ow2.carol.cmi.lb.policy.ILBPolicy;
import org.ow2.carol.cmi.lb.strategy.ILBStrategy;
import org.ow2.carol.cmi.reference.CMIReference;
import org.ow2.carol.cmi.reference.ServerRef;

/**
 * A cache of cluster view at client side.
 * ClientClusterViewManager refreshes periodically this view.
 * @author The new CMI team
 * @see ClientClusterViewManagerImpl
 */
@ThreadSafe
final class ClientClusterViewCache {

    /**
     * Associations object name - replicated data for policy.
     */
    private static final Map<String, LBPolicyData> LBPOLICYDATA =
        new ConcurrentHashMap<String, LBPolicyData>();

    /**
     * Associations object name - name of interface.
     */
    private static final Map<String, String> ITFNAMES =
        new ConcurrentHashMap<String, String>();

    /**
     * Associations object name - replicated.
     */
    private static final Map<String, Boolean> REPLICATED_OBJECTS =
        new ConcurrentHashMap<String, Boolean>();

    /**
     * Associations object name - name of business interface.
     */
    private static final Map<String, String> BUSINESSNAMES =
        new ConcurrentHashMap<String, String>();

    /**
     * Associations object name - type of policy.
     */
    @SuppressWarnings("unchecked")
    private static final Map<String, Class<? extends ILBPolicy>> LBPOLICYCLASSES =
        new ConcurrentHashMap<String, Class<? extends ILBPolicy>>();

    /**
     * Associations object name - type of strategy.
     */
    @SuppressWarnings("unchecked")
    private static final Map<String, Class<? extends ILBStrategy>> LBSTRATEGYCLASS =
        new ConcurrentHashMap<String, Class<? extends ILBStrategy>>();

    /**
     * Associations object name - list of CMI references.
     */
    private static final Map<String, List<CMIReference>> CMIREFERENCES =
        new ConcurrentHashMap<String, List<CMIReference>>();

    /**
     * Associations address - load factor.
     */
    private static final Map<ServerRef, Integer> LOAD_FACTORS =
        new ConcurrentHashMap<ServerRef, Integer>();

    /**
     * Default private constructor.
     */
    private ClientClusterViewCache() {}

    /**
     * Sets a object's load-balancing policy data (maybe a new version).
     * @param objectName a name of object
     * @param lbPolicyData The load balancing policy data
     */
    static void setLBPolicyData(final String objectName, final LBPolicyData lbPolicyData){
        LBPOLICYDATA.put(objectName, lbPolicyData);
    }

    /**
     * @param objectName a name of object
     * @return LB policy data for a given object
     */
    static LBPolicyData getLBPolicyData(final String objectName) {
        return LBPOLICYDATA.get(objectName);
    }

    /**
     * Sets the interface name of the object with the given name.
     * @param objectName a name of object
     * @param itfName an interface name
     */
    static void setItfName(final String objectName, final String itfName){
        ITFNAMES.put(objectName, itfName);
    }

    /**
     * Gets the interface name of the object with the given name.
     * @param objectName The name of object
     * @return the interface name of the object with the given name
     */
    static String getItfName(final String objectName) {
        return ITFNAMES.get(objectName);
    }

    /**
     * Sets the business interface name of the object with the given name.
     * @param objectName a name of object
     * @param businessName a business interface name
     */
    static void setBusinessName(final String objectName, final String businessName){
        BUSINESSNAMES.put(objectName, businessName);
    }

    /**
     * Gets the business interface name of the object with the given name.
     * @param objectName a name of object
     * @return the business interface name of the object with the given name
     */
    public static String getBusinessName(final String objectName) {
        return BUSINESSNAMES.get(objectName);
    }

    /**
     * Gets the class of a load-balancing policy.
     * @param objectName a name of object
     * @return The class of the given load-balancing policy
     */
    @SuppressWarnings("unchecked")
    static Class<? extends ILBPolicy> getLBPolicyClass(final String objectName) {
        return LBPOLICYCLASSES.get(objectName);
    }

    /**
     * Sets the class for the given name of LB strategy.
     * @param objectName a name of object
     * @param lbPolicyClass a class for the given name of LB policy
     */
    @SuppressWarnings("unchecked")
    static void setLBPolicyClass(final String objectName, final Class<? extends ILBPolicy> lbPolicyClass) {
        LBPOLICYCLASSES.put(objectName, lbPolicyClass);
    }

    /**
     * Sets the list of CMIReference of a object with a given name.
     * @param objectName a name of object
     * @param cmiReferences a list of CMIReference
     */
    static void setCMIReferences(final String objectName, final List<CMIReference> cmiReferences) {
        CMIREFERENCES.put(objectName, cmiReferences);
    }

    /**
     * Returns the list of CMIReference of a object with a given name.
     * @param objectName a name of object
     * @return the list of CMIReference of a object with a given name
     */
    static List<CMIReference> getCMIReferences(final String objectName) {
        return CMIREFERENCES.get(objectName);
    }

    /**
     * Returns the class for the given name of LB strategy.
     * @param objectName a name of object
     * @return the class for the given name of LB strategy
     */
    @SuppressWarnings("unchecked")
    static Class<? extends ILBStrategy> getLBStrategyClass(final String objectName) {
        return LBSTRATEGYCLASS.get(objectName);
    }

    /**
     * Sets the class for the given name of LB strategy.
     * @param objectName a name of object
     * @param lbStrategyClass a class for the given name of LB strategy
     */
    @SuppressWarnings("unchecked")
    static void setLBStrategyClass(final String objectName, final Class<? extends ILBStrategy> lbStrategyClass) {
        LBSTRATEGYCLASS.put(objectName, lbStrategyClass);
    }

    /**
     * Returns the load-factor for the server with the given address.
     * @param serverRef a reference on a server
     * @return the load-factor for the server with the given address
     */
    static int getLoadFactor(final ServerRef serverRef) {
        return LOAD_FACTORS.get(serverRef);
    }

    /**
     * Sets the load-factor for the server with the given address.
     * @param serverRef a reference on a server
     * @param loadFactor the load-factor for the server with the given address
     */
    static void setLoadFactor(final ServerRef serverRef, final int loadFactor) {
        LOAD_FACTORS.put(serverRef, loadFactor);
    }

    /**
     * Returns the set of references on known servers.
     * @return the set of references on known servers
     */
    static Set<ServerRef> getAddressesOfServer() {
        return LOAD_FACTORS.keySet();
    }

    /**
     * Return true if the object with the given name is replicated for high-availability.
     * @param objectName a name of object
     * @return true if the object with the given name is replicated for high-availability
     */
    static boolean isReplicated(final String objectName) {
        return REPLICATED_OBJECTS.get(objectName);
    }

    /**
     * Set if the object with the given name is replicated for high-availability.
     * @param objectName a name of object
     * @param replicated true if the object with the given name is replicated for high-availability
     */
    static void setReplicated(final String objectName, final boolean replicated) {
        REPLICATED_OBJECTS.put(objectName, replicated);
    }

}
