/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007 Bull S.A.S.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:ClientClusterViewManager.java 914 2007-05-25 16:48:16Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.carol.cmi.controller.client;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.ejb.EJBHome;
import javax.ejb.EJBObject;

import net.jcip.annotations.ThreadSafe;

import org.ow2.carol.cmi.config.CMIConfig;
import org.ow2.carol.cmi.info.CMIInfoExtractor;
import org.ow2.carol.cmi.info.CMIInfoExtractorException;
import org.ow2.carol.cmi.info.ClusteredObjectInfo;
import org.ow2.carol.cmi.lb.data.LBPolicyData;
import org.ow2.carol.cmi.lb.policy.ILBPolicy;
import org.ow2.carol.cmi.lb.strategy.ILBStrategy;
import org.ow2.carol.cmi.reference.CMIReference;
import org.ow2.carol.cmi.reference.ObjectNotFoundException;
import org.ow2.carol.cmi.reference.ServerRef;
import org.ow2.carol.cmi.rpc.CMIProxy;
import org.ow2.carol.cmi.rpc.CMIProxyFactory;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.carol.cmi.controller.common.AbsClusterViewManager;
import org.ow2.carol.cmi.controller.common.CMIThreadFactory;
import org.ow2.carol.cmi.controller.common.ClusterViewManager;
import org.ow2.carol.cmi.controller.provider.ClientClusterViewProvider;
import org.ow2.carol.cmi.controller.provider.ClusteredClientClusterViewProvider;


/**
 * Manager that contains CMI logic for the client-side.
 * @author The new CMI team
 */
@ThreadSafe
public final class ClientClusterViewManagerImpl extends AbsClusterViewManager implements ClientClusterViewManager {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(ClientClusterViewManagerImpl.class);

    /**
     * The InitialContextFactory name for constructing instances of the protocol-relative context.
     */
    private final String initialContextFactoryName;

    /**
     * The protocol name.
     */
    private final String protocol;

    /**
     * A daemon in charge of updating the cluster view.
     */
    private final Thread clientClusterViewUpdater;

    /**
     * Current selected ClientClusterViewProvider.
     */
    private static volatile ClientClusterViewProvider clientClusterViewProvider;


    /**
     * Builds a new manager for the client-side.
     * @param initialContextFactoryName an name of InitialContextFactory
     * @param protocol The protocol name
     */
    private ClientClusterViewManagerImpl(final String initialContextFactoryName, final String protocol) {
        this.initialContextFactoryName = initialContextFactoryName;
        this.protocol = protocol;

        // Create the daemon in charge of updating the cluster view
        CMIThreadFactory cmiThreadFactory = getCmiThreadFactory();
        clientClusterViewUpdater = cmiThreadFactory.newThread(new Runnable() {
            public void run() {
                // Register this client (for statistics)
                try {
                    clientClusterViewProvider.registerClient(getUUID());
                } catch (RemoteException e) {
                    // No importance !
                    LOGGER.debug("Thread {0}> Cannot register the client", Thread.currentThread().getName(), e);
                }
                do {
                    // Start latter because the system object is already up-to-date
                    int refreshTime = getDelayToRefresh();
                    LOGGER.debug("Thread {0} is sleeping for {1} millis...", Thread.currentThread().getName(), refreshTime);
                    try {
                        Thread.sleep(refreshTime);
                    } catch (InterruptedException e) {
                        LOGGER.debug("Thread client interrupted", e);
                    }
                    LOGGER.debug("Thread {0}> Updating the client cluster view...", Thread.currentThread().getName());
                    try {
                        updateClusterView();
                    } catch (ClientClusterViewManagerException e) {
                        LOGGER.error("Thread {0}> Cannot update the cluster view", Thread.currentThread().getName(), e);
                        throw new ControlThreadException(
                                "Cannot update the cluster view", e);
                    }
                    // Now sleep...
                } while(true);
            }
        });
    }

    /**
     * Static method to get a singleton instance of ClientClusterViewManager.
     * @param initialContextFactoryName a name of class of InitialContextFactory
     * to use for constructing instances of the protocol-relative context
     * @param protocol a name of protocol
     * @param serverRefs a set of references on server
     * @return the instance (singleton) of ClientClusterViewManager
     * @throws ClientClusterViewManagerException if we cannot get an instance of ClientClusterViewManager.
     */
    public static synchronized ClientClusterViewManager getClientClusterViewManager(
            final String initialContextFactoryName, final String protocol, final List<ServerRef> serverRefs)
    throws ClientClusterViewManagerException {

        ClusterViewManager clusterViewManager = AbsClusterViewManager.getClusterViewManager();
        // If none instance of ClusterViewManager exists, we create one
        if(clusterViewManager == null) {
            LOGGER.debug("Constructing a ClientClusterviewManager...");
            ClientClusterViewManagerImpl clientClusterViewManager =
                new ClientClusterViewManagerImpl(initialContextFactoryName, protocol);

            // Initializes data of the provider of cluster view
            clientClusterViewManager.initDataOfProvider(serverRefs);

            // Initializes data of the dummy registry to enable a load-balancing of the JNDI accesses
            clientClusterViewManager.initDataOfDummyRegistry();

            // Start the daemon to update the cluster view
            clientClusterViewManager.clientClusterViewUpdater.start();

            clusterViewManager = clientClusterViewManager;
            AbsClusterViewManager.setClusterViewManager(clusterViewManager);
            LOGGER.info("The client-side manager was successfully started");
        } else {
            if(!(clusterViewManager instanceof ClientClusterViewManager)){
                 LOGGER.error("An instance of the manager that is not a ClientClusterViewManager already exists in the JVM");
                 throw new ClientClusterViewManagerException(
                         "An instance of the manager that is not a ClientClusterViewManager already exists in the JVM");
            }
        }
        return (ClientClusterViewManager) clusterViewManager;
    }

    /**
     * Creates a CMI proxy to enable invocations on the provider of cluster view.
     * @param serverRefs an initial set of the provider URLs for the provider of cluster view
     * @throws ClientClusterViewManagerException if a CMI proxy cannot be obtained for the provider of cluster view
     */
    @SuppressWarnings("unchecked")
    private void initDataOfProvider(final List<ServerRef> serverRefs) throws ClientClusterViewManagerException {
        LOGGER.debug("**** Begin of initialization of the connection at provider...");
        String providerName = CMIConfig.getBindNameForProvider();

        // Initializes the list of provider URL for getting the cluster view
        List<CMIReference> cmiReferences = new ArrayList<CMIReference>();
        for(ServerRef serverRef : serverRefs) {
            cmiReferences.add(new CMIReference(serverRef, providerName));
        }
        ClientClusterViewCache.setCMIReferences(providerName, cmiReferences);

        String itfName = ClientClusterViewProvider.class.getName();

        // Extracts the informations about provider
        ClusteredObjectInfo clusteredObjectInfo;
        try {
            clusteredObjectInfo = CMIInfoExtractor.extractClusteringInfoFromAnnotatedPOJO(
                    itfName, ClusteredClientClusterViewProvider.class, false);
        } catch (CMIInfoExtractorException e) {
            LOGGER.error("Cannot extract the informations about provider", e);
            throw new ClientClusterViewManagerException("Cannot extract the informations about provider", e);
        }

        // Initializes the cache
        ClientClusterViewCache.setItfName(providerName, itfName);
        ClientClusterViewCache.setReplicated(providerName, false);

        String lbPolicyType = clusteredObjectInfo.getPolicyType();
        String lbStrategyType = clusteredObjectInfo.getStrategyType();

        LBPolicyData lbPolicyDataForProvider =
            new LBPolicyData(lbPolicyType, lbStrategyType, clusteredObjectInfo.getProperties());
        ClientClusterViewCache.setLBPolicyData(providerName, lbPolicyDataForProvider);

        try {
            Class<? extends ILBPolicy> lbPolicyClass = (Class<? extends ILBPolicy>) Class.forName(lbPolicyType);
            Class<? extends ILBStrategy> lbStrategyClass = (Class<? extends ILBStrategy>) Class.forName(lbStrategyType);
            ClientClusterViewCache.setLBPolicyClass(providerName, lbPolicyClass);
            ClientClusterViewCache.setLBStrategyClass(providerName, lbStrategyClass);
        } catch (ClassNotFoundException e) {
            LOGGER.fatal("Cannot initialize class to access to the provider", e);
            throw new ClientClusterViewManagerException("Cannot initialize class to access to the provider", e);
        }

        initLocalData(providerName);
        try {
            updateLBPolicy(providerName);
        } catch (ObjectNotFoundException e) {
            LOGGER.error("Cannot initialize informations of the provider", e);
            throw new ClientClusterViewManagerException("Cannot initialize informations of the provider", e);
        }

        // Gets a CMI proxy to enable load-balanced calls at provider
        CMIProxy cmiProxy;
        try {
            cmiProxy = CMIProxyFactory.newCMIProxy(this, providerName, protocol);
        } catch (Exception e) {
            LOGGER.error("Cannot get a CMI proxy for the provider with name {0}", providerName, e);
            throw new ClientClusterViewManagerException("Cannot get a CMI proxy for the provider with name "+providerName, e);
        }
        clientClusterViewProvider = (ClientClusterViewProvider) cmiProxy;
        try {
            updateObject(providerName);
        } catch (ObjectNotFoundException e) {
            throw new ClientClusterViewManagerException("Provider not found", e);
        }
        LOGGER.debug("**** End of initialization of the connection at provider.");
    }

    /**
     * Registers the dummy registry as an object to watch.
     */
    private void initDataOfDummyRegistry() {
        String dummyName = CMIConfig.getBindNameForDummyRegistry();
        initLocalData(dummyName);
        try {
            updateObject(dummyName);
        } catch (ObjectNotFoundException e) {
            throw new ClientClusterViewManagerException("Dummy registry not found", e);
        }
    }

    /**
     * Adds an object with the given name to a set of objects that need to have their state up-to-date.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the name of interface of the object with the given name cannot be downloaded
     */
    public synchronized void addObjectToWatch(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        if(!isLocalDataInitialized(objectName)) {
            LOGGER.debug("Initializing local data for object with name {0}", objectName);
            String itfName;
            try {
                itfName = clientClusterViewProvider.getItfName(objectName);
            } catch (RemoteException e) {
                LOGGER.error("Cannot pull the name of interface for object with name {0}", objectName, e);
                throw new ClientClusterViewManagerException(
                        "Cannot get pull the name of interface for object with name "+objectName, e);
            }
            ClientClusterViewCache.setItfName(objectName, itfName);
            // Check if the object is an ejb2
            if(Arrays.asList(getInterface(objectName).getInterfaces()).contains(EJBHome.class)) {
                String businessName;
                try {
                    businessName = clientClusterViewProvider.getBusinessName(objectName);
                } catch (RemoteException e) {
                    LOGGER.error("Cannot pull the name of business interface for object with name {0}", objectName, e);
                    throw new ClientClusterViewManagerException(
                            "Cannot get pull the name of business interface for object with name "+objectName, e);
                }
                ClientClusterViewCache.setBusinessName(objectName, businessName);
            }
            boolean hasState;
            try {
                hasState = clientClusterViewProvider.isReplicated(objectName);
            } catch (RemoteException e) {
                LOGGER.error("Cannot know if the object with name {0} has a state", objectName, e);
                throw new ClientClusterViewManagerException(
                        "Cannot know if the object with name "+objectName+" has a state", e);
            }
            ClientClusterViewCache.setReplicated(objectName, hasState);
            initLocalData(objectName);
        }
    }

    /**
     * Returns the interface of an object bound with the given name.
     * @param objectName a name of object
     * @return the interface of an object bound with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the class cannot be loaded
     */
    public Class<?> getInterface(final String objectName) throws ObjectNotFoundException, ClientClusterViewManagerException {
        String itfName = ClientClusterViewCache.getItfName(objectName);
        if(itfName == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        try {
            return Class.forName(itfName);
        } catch (ClassNotFoundException e) {
            throw new ClientClusterViewManagerException("Cannot load the interface for name "+itfName, e);
        }
    }

    /**
     * Returns the interface of an object bound with the given name.
     * @param objectName a name of object
     * @return the interface of an object bound with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the class cannot be loaded
     */
    @SuppressWarnings("unchecked")
    public Class<? extends EJBObject> getRemoteClass(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        String businessName = ClientClusterViewCache.getBusinessName(objectName);
        if(businessName == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        try {
            return (Class<? extends EJBObject>) Class.forName(businessName);
        } catch (ClassNotFoundException e) {
            throw new ClientClusterViewManagerException("Cannot load the business interface for name "+businessName, e);
        }
    }

    /**
     * Returns the class that defines the policy for an object with the given name.
     * @param objectName a name of object
     * @return the class that defines the policy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the class of policy of the object with the given name cannot be downloaded
     */
    @SuppressWarnings("unchecked")
    public Class<? extends ILBPolicy> getLBPolicyClass(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        Class<? extends ILBPolicy> lbPolicy = ClientClusterViewCache.getLBPolicyClass(objectName);
        if(lbPolicy == null) {
            try {
                lbPolicy = pullLBPolicyClass(objectName);
            } catch (RemoteException e) {
                LOGGER.error("Cannot set the class of policy for {0}", objectName, e);
                throw new ClientClusterViewManagerException("Cannot set the class of policy for "+objectName, e);
            }
            ClientClusterViewCache.setLBPolicyClass(objectName, lbPolicy);
        }
        return lbPolicy;
    }

    /**
     * Returns the class that defines the strategy for an object with the given name.
     * @param objectName a name of object
     * @return the class that defines the strategy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the class of strategy of the object with the given name cannot be downloaded
     */
    @SuppressWarnings("unchecked")
    public Class<? extends ILBStrategy> getLBStrategyClass(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        Class<? extends ILBStrategy> lbStrategy = ClientClusterViewCache.getLBStrategyClass(objectName);
        if(lbStrategy == null) {
            try {
                lbStrategy = pullLBStrategyClass(objectName);
            } catch (RemoteException e) {
                LOGGER.error("Cannot set the class of strategy for {0}", objectName, e);
                throw new ClientClusterViewManagerException("Cannot set the class of strategy for "+objectName, e);
            }
            ClientClusterViewCache.setLBStrategyClass(objectName, lbStrategy);
        }
        return lbStrategy;
    }

    /**
     * Returns a list of CMIReference for an object with the given name and protocol.
     * @param objectName a name of object
     * @param protocolName a name of protocol
     * @return a list of CMIReference for an object with the given name and protocol
     * @throws ObjectNotFoundException if none object has the given name for the given protocol
     */
    public List<CMIReference> getCMIReferences(final String objectName, final String protocolName)
    throws ObjectNotFoundException {
        List<CMIReference> cmiReferences = ClientClusterViewCache.getCMIReferences(objectName);
        if(cmiReferences == null) {
            LOGGER.debug("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        return cmiReferences;
    }

    /**
     * Returns the date of properties for an object with the given name.
     * @param objectName a name of object
     * @return the date of properties for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public long getDateOfProperties(final String objectName) throws ObjectNotFoundException {
        LBPolicyData lbPolicyData = ClientClusterViewCache.getLBPolicyData(objectName);
        if(lbPolicyData == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        return lbPolicyData.getDateOfProperties();
    }

    /************* Start of methods to pull the cluster view: *****************/

    /**
     * Pulls the Load Balance policy class.
     * @param objectName a name of object
     * @return A Load Balance policy class
     * @throws ObjectNotFoundException if none object has the given name
     * @throws RemoteException if the class of policy of the object with the given name cannot be downloaded
     */
    @SuppressWarnings("unchecked")
    private Class<? extends ILBPolicy> pullLBPolicyClass(final String objectName)
    throws ObjectNotFoundException, RemoteException {
        // Set classloader to download class of policy
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new CMIClientClassLoader(oldClassLoader, this));
        Class<? extends ILBPolicy> lbPolicyClass;
        try {
            lbPolicyClass = clientClusterViewProvider.getLBPolicyClass(objectName);
        } catch(ClassNotFoundException e) {
            LOGGER.error("Cannot download the LB policy class for {0}", objectName, e);
            throw new ClientClusterViewManagerException("Cannot download the LB policy class for " + objectName, e);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
        return lbPolicyClass;
    }

    /**
     * Pulls the Load Balance strategy class.
     * @param objectName a name of object
     * @return A Load Balance strategy class
     * @throws ObjectNotFoundException if none object has the given name
     * @throws RemoteException if the class of strategy of the object with the given name cannot be downloaded
     */
    @SuppressWarnings("unchecked")
    private Class<? extends ILBStrategy> pullLBStrategyClass(final String objectName)
    throws RemoteException, ObjectNotFoundException {
        // Set classloader to download class of strategy
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(new CMIClientClassLoader(oldClassLoader, this));
        Class<? extends ILBStrategy> lbStrategyClass;
        try {
            lbStrategyClass = clientClusterViewProvider.getLBStrategyClass(objectName);
        } catch(ClassNotFoundException e) {
            LOGGER.error("Cannot download the LB policy class for {0}", objectName, e);
            throw new ClientClusterViewManagerException("Cannot download the LB strategy class for " + objectName, e);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }
        return lbStrategyClass;
    }

    /**
     * Returns true if the pool for object with the given name should be empty.
     * @param objectName a name of object
     * @return true if the pool for object with the given name should be empty
     * @throws ObjectNotFoundException if no object is bound with the given name
     * @throws ClientClusterViewManagerException if the flag to empty pool for the object with the given name cannot be downloaded
     */
    public boolean isPoolToEmpty(final String objectName) throws ObjectNotFoundException, ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.isPoolToEmpty(objectName);
        } catch (RemoteException e) {
            LOGGER.error("Cannot pull the flag to reset or no the pool for object {0}", objectName, e);
            throw new ClientClusterViewManagerException(
                    "Cannot pull the flag to reset or no the pool for object "+objectName, e);
        }
    }

    /************* End of methods to pull the cluster view *****************/

    /**
     * Returns the properties of the LB policy for an object with the given name.
     * @param objectName a name of object
     * @return the properties of the LB policy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Map<String, Object> getPropertiesForLBPolicy(final String objectName) throws ObjectNotFoundException {
        LBPolicyData lbPolicyData = ClientClusterViewCache.getLBPolicyData(objectName);
        if(lbPolicyData == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        return lbPolicyData.getProperties();
    }

    /**
     * Returns a value property of the LB policy for an object with the given name.
     * @param objectName a name of object
     * @param propertyName a name of property
     * @return a value property of the LB policy for an object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public Object getPropertyForLBPolicy(final String objectName, final String propertyName)
    throws ObjectNotFoundException {
        LBPolicyData lbPolicyData = ClientClusterViewCache.getLBPolicyData(objectName);
        if(lbPolicyData == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        return lbPolicyData.getProperties().get(propertyName);
    }

    /**
     * Returns a name of class that implements the interface InitialContextFactory for the given name of protocol.
     * @param protocolName a name of protocol
     * @return a name of class that implements the interface InitialContextFactory for the given name of protocol
     */
    public String getInitialContextFactoryName(final String protocolName) {
        return initialContextFactoryName;
    }

    /**
     * Returns the protocol used by this manager.
     * @return the protocol used by this manager
     */
    public String getProtocol() {
        return protocol;
    }

    /**
     * Returns the bytecode of the class with the given binary name.
     * @param binaryName a binary name of a class
     * @return the bytecode of the class with the given binary name
     * @throws ClientClusterViewManagerException if the bytecode cannot obtained
     */
    public byte[] getBytecode(final String binaryName) throws ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.getBytecode(binaryName);
        } catch(Exception e) {
            LOGGER.error("Cannot get bytecode for {0}", binaryName, e);
            throw new ClientClusterViewManagerException("Cannot get bytecode for "+binaryName, e);
        }
    }

    /**
     * Returns true if the object with the specified name is clustered.
     * @param objectName a name of object
     * @return true if the object with the specified name is clustered
     * @throws ClientClusterViewManagerException if the flag to indicate if the object, with the given name,
     * is clustered cannot be downloaded
     */
    public boolean isClustered(final String objectName) throws ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.isClustered(objectName);
        } catch (RemoteException e) {
            LOGGER.error("Connection has failed", e);
            throw new ClientClusterViewManagerException("Connection has failed", e);
        }
    }

    /**
     * Returns the time between each update of the cluster view by clients.
     * @return the time between each update of the cluster view by clients
     * @throws ClientClusterViewManagerException if the delay between each update of the cluster view cannot be donwload
     */
    public int getDelayToRefresh() throws ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.getDelayToRefresh();
        } catch (RemoteException e) {
            LOGGER.error("Cannot get delay to refresh", e);
            throw new ClientClusterViewManagerException("Cannot get delay to refresh", e);
        }
    }

    /**
     * Returns the name of class of strategy for the object with the given name.
     * @param objectName a name of object
     * @return the name of class of strategy for the object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public String getLBStrategyClassName(final String objectName) throws ObjectNotFoundException {
        LBPolicyData lbPolicyData = ClientClusterViewCache.getLBPolicyData(objectName);
        if(lbPolicyData == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        return lbPolicyData.getStrategyType();
    }

    /**
     * Returns the name of class of policy for the object with the given name.
     * @param objectName a name of object
     * @return the name of class of policy for the object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public String getLBPolicyClassName(final String objectName) throws ObjectNotFoundException {
        LBPolicyData lbPolicyData = ClientClusterViewCache.getLBPolicyData(objectName);
        if(lbPolicyData == null) {
            LOGGER.error("Object is unknown: {0}", objectName);
            throw new ObjectNotFoundException("Object is unknown: "+objectName);
        }
        return lbPolicyData.getPolicyType();
    }

    /**
     * Returns the name of cluster for the object with the given name.
     * @param objectName a name of object
     * @return the name of cluster for a object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the name of the cluster, that contains the object with the given name,
     * cannot be downloaded
     */
    public String getClusterName(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        try {
            return clientClusterViewProvider.getClusterName(objectName);
        } catch (RemoteException e) {
            LOGGER.error("Cannot get name of cluster for object with name {0}", objectName, e);
            throw new ClientClusterViewManagerException("Cannot get name of cluster for object with name "+objectName, e);
        }
    }

    /**
     * Updates informations about clustering for the object with the given name.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the update failed
     */
    public void updateObject(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException {
        LOGGER.debug("Updating infos for: {0} - thread is {1}", objectName, Thread.currentThread().getName());
        try {
            pullAndUpdateLBPolicy(objectName);
            pullAndUpdateCMIReferences(objectName);
            pullAndUpdatePoolSize(objectName);
        } catch (Exception e) {
            LOGGER.error("Cannot pull the informations for object with name {0}", objectName, e);
            throw new ClientClusterViewManagerException("Cannot pull the informations for object with name "+objectName, e);
        }
    }

    /**
     * Updates informations about clustering for every watched objects.
     * @throws ClientClusterViewManagerException if the update failed
     */
    public void updateClusterView() throws ClientClusterViewManagerException {
        // Update list of node for each clustered RMI object
        for(String objectName : getNamesOfInitializedObjects()) {
            try {
                updateObject(objectName);
            } catch (ObjectNotFoundException e) {
                LOGGER.fatal("The object with name {0} is unknown", objectName, e);
                throw new ClientClusterViewManagerException("The object with name "+objectName+" is unknown", e);
            }
        }
        pullAndUpdateLoadFactors();
    }

    /**
     * Pulls and updates the policy for the object with the given name.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if some informations miss or the new policy cannot be constructed
     * @throws RemoteException if the information about policy cannot be downloaded
     */
    private void pullAndUpdateLBPolicy(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException, RemoteException {

        // Pulls the name of class for policy
        String lastLBPolicyName = clientClusterViewProvider.getLBPolicyClassName(objectName);
        if(lastLBPolicyName == null) {
            LOGGER.error("The new name of LB policy is null.");
            throw new ClientClusterViewManagerException("The new name of LB policy is null.");
        }
        LOGGER.debug("New LB policy name: {0}", lastLBPolicyName);

        // Pulls the name of class for strategy
        String lastLBStrategyName = clientClusterViewProvider.getLBStrategyClassName(objectName);
        if(lastLBStrategyName == null) {
            LOGGER.error("The new name of LB strategy is null.");
            throw new ClientClusterViewManagerException("The new name of LB strategy is null.");
        }
        LOGGER.debug("New LB strategy name: {0}", lastLBStrategyName);

        // Pulls the date of properties
        long dateOfLastProps = clientClusterViewProvider.getDateOfProperties(objectName);
        LOGGER.debug("New date of properties: {0}", dateOfLastProps);

        // Compares with the current data
        if(ClientClusterViewCache.getLBPolicyData(objectName) != null) {
            // Cache contains some data
            String lbPolicyName = getLBPolicyClassName(objectName);
            LOGGER.debug("Old LB policy name: {0}", lbPolicyName);
            long dateOfProps = getDateOfProperties(objectName);
            LOGGER.debug("Old date of properties: {0}", dateOfProps);
            String lbStrategyName = getLBStrategyClassName(objectName);
            LOGGER.debug("Old LB strategy name: {0}", lbStrategyName);
            if(lbPolicyName.equals(lastLBPolicyName)
                    && dateOfProps == dateOfLastProps
                    && lbStrategyName.equals(lastLBStrategyName)) {
                return;
            }
        }
        Map<String, Object> lastProperties;

        if(dateOfLastProps == 0) {
            // No property
            lastProperties = new HashMap<String, Object>();
        } else {
            lastProperties = clientClusterViewProvider.getPropertiesForLBPolicy(objectName);
        }
        ClientClusterViewCache.setLBPolicyData(objectName, new LBPolicyData(
                lastLBPolicyName, lastLBStrategyName, lastProperties, dateOfLastProps));
        try {
            updateLBPolicy(objectName);
        } catch (ObjectNotFoundException e) {
            LOGGER.error("Cannot get LB policy for object {0}", objectName, e);
            throw new ClientClusterViewManagerException("Cannot get LB policy for object "+objectName, e);
        }
    }

    /**
     * Pulls and updates the CMI references for the object with the given name.
     * @param objectName a name of object
     * @throws ObjectNotFoundException if none object has the given name
     * @throws ClientClusterViewManagerException if the downloaded list is null
     * @throws RemoteException if the list cannot be downloaded
     */
    private void pullAndUpdateCMIReferences(final String objectName)
    throws ObjectNotFoundException, ClientClusterViewManagerException, RemoteException {
        //Update the list of nodes for this RMI object
        List<CMIReference> cmiReferences = clientClusterViewProvider.getCMIReferences(objectName, protocol);
        if(cmiReferences == null) {
            LOGGER.error("The new node list for {0} is null", objectName);
            throw new ClientClusterViewManagerException("The new node list for "+objectName+" is null");
        }
        ClientClusterViewCache.setCMIReferences(objectName, cmiReferences);
    }

    /**
     * Pulls and updates the configuration of the pool for the object with the given name.
     * @param objectName a name of object
     */
    private void pullAndUpdatePoolSize(final String objectName) {
        // TODO Auto-generated method stub

    }

    /**
     * Pulls and updates the load-factors of the server known by this client.
     */
    private void pullAndUpdateLoadFactors() {
        for(ServerRef address : ClientClusterViewCache.getAddressesOfServer()) {
            int loadFactor;
            try {
                loadFactor = clientClusterViewProvider.getLoadFactor(address);
                ClientClusterViewCache.setLoadFactor(address, loadFactor);
            } catch (Exception e) {
                LOGGER.error("Cannot pull the load factor for server with address {0}", address, e);
                throw new ClientClusterViewManagerException(
                        "Cannot pull the load factor for server with address "+address, e);
            }
        }
    }

    /**
     * Returns the maximal size of pool of CMIReferenceable for a object with the given name.
     * @param objectName a name of object
     * @return the maximal size of pool of CMIReferenceable for a object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public int getMaxPoolSize(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewProvider.getMaxPoolSize(objectName);
        } catch (RemoteException e) {
            LOGGER.error("Cannot pull the maximal size of pool for object with name {0}", objectName, e);
            throw new ClientClusterViewManagerException(
                    "Cannot pull the maximal size of pool for object with name "+objectName, e);
        }
    }

    /**
     * Returns the minimal size of pool of CMIReferenceable for a object with the given name.
     * @param objectName a name of object
     * @return the minimal size of pool of CMIReferenceable for a object with the given name
     * @throws ObjectNotFoundException if none object has the given name
     */
    public int getMinPoolSize(final String objectName) throws ObjectNotFoundException {
        try {
            return clientClusterViewProvider.getMinPoolSize(objectName);
        } catch (RemoteException e) {
            LOGGER.error("Cannot pull the minimal size of pool for object with name {0}", objectName, e);
            throw new ClientClusterViewManagerException(
                    "Cannot pull the minimal size of pool for object with name "+objectName, e);
        }
    }

    /**
     * Returns the load-factor for the server with the given address.
     * @param serverRef a reference on a server
     * @return the load-factor for the server with the given address
     */
    public int getLoadFactor(final ServerRef serverRef) {
        return ClientClusterViewCache.getLoadFactor(serverRef);
    }

    /**
     * Return true if the object with the given name is replicated for high-availability.
     * The module 'ha' is so required.
     * @param objectName a name of object
     * @return true if the object with the given name is replicated for high-availability
     * @throws ObjectNotFoundException if none object has the given name
     */
    public boolean isReplicated(final String objectName)
    throws ObjectNotFoundException {
        return ClientClusterViewCache.isReplicated(objectName);
    }

}
